# -*- coding: utf-8 -*-
#
# “My Hovercraft is Full of Eels” translations
#
# Taken from
#   "Omniglot - writing systems and languages of the world".
#   Copyright (c) 1998-2018 Simon Ager. All rights reserved.
#
# Redistributed for non-commercial purposes under the copyright notice
# offered by Simon Ager on the omniglot.com web site:
#
#   “If you would like to use the material for personal, educational or
#   non-commercial purposes you're welcome to do so. Don't forget to mention
#   where it came from.”
#
# For the original compilation of translations,
# see <http://www.omniglot.com/language/phrases/hovercraft.htm>.
#
# For the copyright notice allowing distribution for non-commercial purposes,
# see <http://www.omniglot.com/copyright.htm>.
#
#
# Separated as file with nonfree material, because the author of TypeAtlas is
# not a lawyer, and cannot make interpretations of copyright law beyond
# the established license or notices, presented above. This separation itself
# does not constitute a claim that the file is indeed non-free, should it be
# used in jurisdictions where data or a database of user-submitted phrases
# (in the manner used by TypeAtlas) *hypotethically* might be
# distributable under terms satisfying The Free Software Definition. For any use
# of this file beyond the scope of the notice from Simon Ager above, consult a
# lawyer.
#

HOVERCRAFT_ENGLISH = 'My hovercraft is full of eels'
HOVERCRAFT_TRANSLATION_SOURCE = 'http://www.omniglot.com/language/phrases/hovercraft.htm'
HOVERCRAFT_NOTE = 'The phrase originated in Monty Python'

hovercraft = {
    'en': HOVERCRAFT_ENGLISH,
    'af': 'My skeertuig is vol palings',
    'sq-alt': 'Anija jêm ãsht plot mê ngjala',
    'sq': 'Automjeti im është plot me ngjala',
    'am': 'የኔ ማንዣበቢያ መኪና በእባቦች ተሞልቷል',
    'ar': 'حَوّامتي مُمْتِلئة بِأَنْقَلَيْسون',
    'an': "O mío aeroeslisador ye pleno d'anguilas",
    'hy': 'Իմ օդաթիռը լի է օձաձկերով',
    'hyw': ' Իմ սաւառնակս օձաձկներով լեցուն է:',
    'as': "মোৰ হ'ভাৰক্ৰাফ্টটো ঈল মাছেৰে ভৰি গৈছে",
    'az': 'Hoverkraftimin içi ilan balıǧı ilə doludur',
    'eu': 'Nire aerolabaingailua aingirez beteta dago',
    'be': 'Маё судна на паветранай падушцы поўна вуграмі',
    'bn': 'আমার হভারক্রাফ্ট কুঁচে মাছ-এ ভরা হয়ে গেছে',
    'bs': 'Moja lebdjelica je puna jegulja',
    'br': 'Leun gant sili eo ma dourruzer',
    'bg': 'Корабът ми на въздушна възглавница е пълен със змиорки',
    'my': 'ကျွန်တော်ရဲ့ လေစီးယာဉ်မှာ ငါးရှင့်တွေအပြည့်ရှိနေပါတယ်။',
    'ca': "El meu aerolliscador està ple d'anguiles",
    'ch': 'Bula titugi eyo-ku hovercraft',
    'ch-gu': 'Bula titugi eyo-ku hovercraft',
    'ch-mp': 'Bula hasuli gi halom hovercraft-hu',
    'zh-yue': '(Cantonese) 我隻氣墊船裝滿晒鱔',
    'zh': '我的氣墊船裝滿了鱔魚 [我的气垫船装满了鳝鱼]',
    'zh-tw': '我的氣墊船充滿了鰻魚',
    'kw': 'Leun a sylli yw ow skath bargesi',
    'co': "U me battellu hè carcu d'anguili",
    'hr': 'Moja je lebdjelica puna jegulja',
    'cs': 'Moje vznášedlo je plné úhořů',
    'da': 'Mit luftpudefartøj er fyldt med ål',
    'nl': 'Mijn luchtkussenboot zit vol paling',
    'et': 'Mu hõljuk on angerjaid täis',
    'fo': "Luftpútufarið hjá mær yður í álli",
    'fj': 'Na noqu hovercraft e sinai tu vei na duna',
    'fi': 'Ilmatyynyalukseni on täynnä ankeriaita',
    'fr': "Mon aéroglisseur est plein d'anguilles",
    'fy': 'Myn hovercraft sit fol mei iel',
    'frr': 'Min luftdümpetbüüdj as ful ma äil',
    'gl': 'O meu aerodeslizador esta cheo de anguías',
    'ka': 'ჩემი ხომალდი საჰაერო ბალიშზე სავსეა გველთევზებით',
    'de': 'Mein Luftkissenfahrzeug ist voller Aale',
    'el-ancient': 'Τὸ ἐμὸν αερόστρωμνον ἐγχελείων πλῆρές ἐστιν',
    'el': 'Το αερόστρωμνο μου ειναι γεμάτο χέλια',
    'gu': 'મારું હોવેરક્રાફ્ટ ઈલ માછલીઓ થી ભરાઈ ગયું છે',
    'ha': 'Jirgina a cike yake da bano',
    'haw': 'Pihaʻū oʻu mokukauaheahe i nā puhi',
    'he': 'הרחפת שלי מלאה בצלופחים',
    'hi': 'मेरी मँडराने वाली नाव सर्पमीनों से भरी हैं',
    'hu': 'A légpárnás hajóm tele van angolnákkal',
    'is': 'Svifnökkvinn minn er fullur af álum',
    'ig': 'Hovercraft m juputara azu arira',
    'smn': 'Muu áimukuáddákäärbis lii tievâ aŋgeriaid',
    'id': 'Hovercraft saya penuh dengan belut',
    'iu': 'ᐅᒥᐊᕐᔫᑉ ᐳᓪᓕᓕᒫᐸᒐ ᑕᑦᑕᕐᓂᖅ ᐊᒻᒪᔭᖅ',
    'ga': "Tá m'árthach foluaineach lán d'eascanna",
    'it': 'Il mio aeroscivolante è pieno di anguille',
    'ja': '私のホバークラフトは鰻でいっぱいです',
    'kn': 'ನನ್ನ ಯಾಂತ್ರಿಕದೋಣಿ ಹಾವುಮೀನುಗಳಿಂದ ತುಂಬಿದೆ',
    'ks': 'Myon hovercraaft chhu eelav seeth bharith',
    'kk': 'Менің әуе негіз кемесi жыланбалықпен толтырылған',
    'km': 'សុទ្ធ​តែ​អន្ទង់​ពេញ​ទូក​ហោះ​យើង',
    'ki': 'E onrake au kaibuke n te rabono',
    'ko': '내 호버크라프트는 장어로 가득 차 있어요',
    'ku-alt': 'Hoverkrafta mi(n) tijé marmasî e',
    'ku': 'Seyarey balindîyekem pirr ye li marmasî',
    'ky': 'Менин аба кемем курт балыктар менен толуп турат',
    'lo': 'ມີປາໄຫຼເຕັມຢູ່ໃນເຮືອພັດລົມຂອງຂ້ອຍ',
    'la-alt': 'Mea navis volitans anguillis plena est',
    'la': 'Navis volitans mihi anguillis plena est',
    'lv': 'Mans transportlīdzeklis uz gaisa spilvena ir pilns ar zušiem',
    'lt': 'Lithuanian  Mano laivas su oro pagalve pilnas ungurių',
    'nds': 'Mien Lüchtpöukaan ist vull von sük aalen',
    'lb': 'Mäi Loftkësseboot ass voller Éilen',
    'mk': 'Моето летачко возило е полно со јагули',
    'ms': 'Hoverkraf saya penuh dengan belut',
    'ml': 'എന്‍റെ പറക്കും-പേടകം നിറയെ വ്ളാങ്കുകളാണ്',
    'mt': 'Il-hovercraft tiegħi hu mimli sallur',
    'gv-alt': 'Ta my haagh crowal lane dy astanyn',
    'gv': 'Ta my lhong chrowal lane dy astanyn',
    'mi': 'Kī tōnu taku waka topaki i te tuna',
    'mr': 'माझी होडी मासळयांनी भरली आहे',
    'mn': 'Миний хөвөгч онгоц могой загасаар дүүрсэн байна',
    'nd': 'Isikepe sami sigcwele umzowa',
    'se': 'Mu áibmofatnasis lea dievva ággarasaid',
    'no-alt': 'Det er fullt av ål i luftputebåten min',
    'no-alt': 'Luftputebåten min er full av ål',
    'no': 'Luftputefartøyet mitt er fullt av ål',
    'oc': "Mon aerolisador es plen d'anguilas",
    'or': 'ମୋ ହୋବର୍କ୍ରାଫ୍ଟ ରେ ଇଲ୍ ଭର୍ତି ହେଇ ଜାଇଛି।',
    'om': 'Koonkalata baariisu ko bofa bishaani gutu',
    'fa': 'هاورکرافت من پر مارماهى است',
    'pl': 'Mój poduszkowiec jest pełen węgorzy',
    'pt': 'O meu hovercraft está cheio de enguias',
    'pt-br': 'Meu aerodeslizador está cheio de enguias',
    'pt-br-alt': 'Meu aerobarco está cheio de enguias',
    'ro': 'Aeroglisorul meu e plin de țipari',
    'ru': 'Моё су́дно на возду́шной поду́шке по́лно угре́й',
    'sm': "Ua tumu la'u ato fagota i pusi",
    'sc': 'Su hovercraft est penu de ambidas',
    'sr': 'Мој ховеркрафт је пун јегуља',
    'sn': 'Hovercraft yangu yakazara nemikunga',
    'si': 'මාගේ වායු පා යානයේ ආඳන් පිරී ඇත',
    'si-alt': 'මගේ වායු පා යානෙ ආඳො පරීලා',
    'sk': 'Moje vznášadlo je plné úhorov',
    'sl': 'Moje vozilo na zračni blazini je polno jegulj',
    'so': 'Huufarkarafkayga waxaa ka buuxa eels',
    'es': 'Mi aerodeslizador está lleno de anguilas',
    'su': 'Kapal ngalayang abdi pinuh ku belut',
    'sw': 'Gari langu linaloangama limejaa na mikunga',
    'sv': 'Min svävare är full med ålar',
    'tl': 'Puno ng palos ang aking hoberkrap',
    'ta-alt': 'என் மிதவை நிறைய விலாங்கு மீன்கள்',
    'ta': 'என் விமானம் நிறைய விலாங்கு மீன்கள்',
    'te': 'నా విమానము అంతా మలుగు చేపలతో నిండిపోయింది',
    'th': 'โฮเวอร์คราฟท์ของผมเต็มไปด้วยปลาไหล',
    'th-alt': 'โฮเวอร์คราฟท์ของฉันเต็มไปด้วยปลาไหล',
    'bo': 'ངའི་རླུང་འདེགས་གྲུ་གཟིངས་ནི་ཉ་མན་ལི་ཡིས་གང་གི་འདུག །',
    'ti': 'ናተይ ሆበርክራፍት ሙሉእ እዩ ብዓሣ',
    'to': "Hoku vakapuna tētē 'i 'olunga fono 'i he toke.",
    'tn': 'Mokoro wa me o tletse ditlhapi',
    'tr-alt': 'Hoverkraftım yılan balığı dolu',
    'tr': 'Hoverkraftımın içi yılan balığı dolu',
    'tvl': 'Taku hovercraft kō fonu i pusi.',
    'uk': 'Моє судно на повітряній подушці наповнене вуграми',
    'uz': "Mening havo yostiqli kemam ilonbalig'i bilan to'lgan",
    'vi-alt': 'Tàu cánh ngầm của tôi đầy lươn',
    'vi': 'Nói một thứ tiếng thì không bao giờ đủ',
    'cy': 'Mae fy hofrenfad yn llawn llyswennod',
    'wo': "Wolof   Sama aéroglisseur fèèss neu ak anguilles'",
    'yi': 'מײַן שוועבשיף איז פֿול מיט ווענגערס',
    'yi-alt': 'מײַן פּראָם איז פֿול מיט ווענגערס',
    'xh': 'Inqwelo etshitshiliza phezu kwamanzi izele ziipalanga',
    'yo': 'Ọkọ afategun-sare mi kun fun ẹja arọ',
    'zu': 'Umkhumbi wami ugcwele ngenyoka zemanzini',
}